item_types = {
    "Dashboard": ["Dashboard", "dashboards"],
    "DataPipeline": ["Data Pipeline", "dataPipelines", "pipeline-content.json"],
    "Datamart": ["Datamart", "datamarts"],
    "Environment": ["Environment", "environments"],
    "Eventhouse": ["Eventhouse", "eventhouses", "EventhouseProperties.json"],
    "Eventstream": ["Eventstream", "eventstreams", "eventstream.json"],
    "GraphQLApi": ["GraphQL Api", "GraphQLApis"],
    "KQLDashboard": ["KQL Dashboard", "kqlDashboards", "RealTimeDashboard.json"],
    "KQLDatabase": [
        "KQL Database",
        "kqlDatabases",
    ],  # "DatabaseProperties.json", "DatabaseSchema.kql"
    "KQLQueryset": ["KQL Queryset", "kqlQuerysets", "RealTimeQueryset.json"],
    "Lakehouse": ["Lakehouse", "lakehouses"],
    "MLExperiment": ["ML Experiment", "mlExperiments"],
    "MLModel": ["ML Model", "mlModels"],
    "MirroredDatabase": [
        "Mirrored Database",
        "mirroredDatabases",
        "mirroredDatabase.json",
    ],
    "MirroredWarehouse": ["Mirrored Warehouse", "mirroredWarehouses"],
    "MountedDataFactory": [
        "Mounted Data Factory",
        "mountedDataFactories",
        "mountedDataFactory-content.json",
    ],
    "Notebook": ["Notebook", "notebooks"],
    "PaginatedReport": ["Paginated Report", "paginatedReports"],
    "Reflex": ["Reflex", "reflexes", "ReflexEntities.json"],
    "Report": ["Report", "reports", "report.json"],
    "SQLDatabase": ["SQL Database", "sqlDatabases"],
    "SQLEndpoint": ["SQL Endpoint", "sqlEndpoints"],
    "SemanticModel": ["Semantic Model", "semanticModels", "model.bim"],
    "SparkJobDefinition": [
        "Spark Job Definition",
        "sparkJobDefinitions",
        "SparkJobDefinitionV1.json",
    ],
    "Warehouse": ["Warehouse", "warehouses"],
}


items = {
    "CopyJob": "copyJobs",
    "Dataflow": "dataflows",
    "Eventhouse": "eventhouses",
    "GraphQLApi": "GraphQLApis",
    "Report": "reports",
    "SemanticModel": "semanticModels",
    # "Environment": "environments",
    "KQLDatabase": "kqlDatabases",
    "KQLDashboard": "kqlDashboards",
    "KQLQueryset": "kqlQuerysets",
    "DataPipeline": "dataPipelines",
    "Notebook": "notebooks",
    "SparkJobDefinition": "sparkJobDefinitions",
    "Eventstream": "eventstreams",
    "MirroredWarehouse": "mirroredWarehouses",
    "MirroredDatabase": "mirroredDatabases",
    "MountedDataFactory": "mountedDataFactories",
    "VariableLibrary": "variableLibraries",
    "ApacheAirFlowJob": "ApacheAirflowJobs",
    "WarehouseSnapshot": "warehousesnapshots",
    "DigitalTwinBuilder": "digitaltwinbuilders",
    "DigitalTwinBuilderFlow": "DigitalTwinBuilderFlows",
    "MirroredAzureDatabricksCatalog": "mirroredAzureDatabricksCatalogs",
    "Map": "Maps",
    "AnomalyDetector": "anomalydetectors",
}
